// ------------------------------------------------------------------------
/*
  Código em SYCL para o cálculo paralelo heterogêneo (CPU + GPU) de:
                              ___
                x = 3 * x - \| x  ,

  onde x é um vetor.
  
  A multiplicação e raiz quadrada são processadas concorrentemente, onde a CPU
  calcula paralelamente '3*x' e a GPU 'sqrt(x)'. Finalmente, apenas a CPU faz,
  em paralelo, a subtração dos resultados parciais e atribuição final.
*/
// ------------------------------------------------------------------------
#include "CL/sycl.hpp"
#include <iostream>
#include <math.h>
#include <array>
using namespace sycl;
constexpr int N = 42; 

int main() {

  std::array<float,N> X, cpuY, gpuY;
  for (int i = 0; i < N; i++) X[i] = i;

  //for( int i = 0; i < N; ++i ) std::cout << '[' << 3 * X[i] - sqrt(X[i]) << ']'; std::cout << std::endl;

  queue cpu_fila{cpu_selector{}};
  queue gpu_fila{gpu_selector{}};

  std::cout << "Dispositivo CPU: " << cpu_fila.get_device().get_info<info::device::name>()   << "\n";
  std::cout << "                 " << cpu_fila.get_device().get_info<info::device::vendor>() << "\n";

  std::cout << "Dispositivo GPU: " << gpu_fila.get_device().get_info<info::device::name>()   << "\n";
  std::cout << "                 " << gpu_fila.get_device().get_info<info::device::vendor>() << "\n";

  {
     buffer bufferX{X};
     buffer cpu_bufferY{cpuY};
     buffer gpu_bufferY{gpuY};
  
     cpu_fila.submit([&](handler &h) {
       accessor x{bufferX, h, read_only};
       accessor cpu_y{cpu_bufferY, h, write_only};
       h.parallel_for(N, [=](id<1> i) {
         cpu_y[i] = 3 * x[i];
       });
     });
  
     gpu_fila.submit([&](handler &h) {
       accessor x{bufferX, h, read_only};
       accessor gpu_y{gpu_bufferY, h, write_only};
       h.parallel_for(N, [=](id<1> i) {
         gpu_y[i] = sqrt(x[i]);
       });
     });
  
     cpu_fila.submit([&](handler &h) {
       accessor x{bufferX, h, write_only};
       accessor cpu_y{cpu_bufferY, h, read_only};
       accessor gpu_y{gpu_bufferY, h, read_only};
       h.parallel_for(N, [=](id<1> i) {
         x[i] = cpu_y[i] - gpu_y[i];
       });
     });
  }

  for( int i = 0; i < N; ++i ) std::cout << '[' << X[i] << ']'; std::cout << std::endl;
  return 0;
}
